/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.gui.quickview;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.TransferHandler;

import ca.uvic.csr.shrimp.ShrimpApplication.ApplicationAccessor;
import ca.uvic.csr.shrimp.ShrimpProject.ShrimpProject;
import ca.uvic.csr.shrimp.resource.ResourceHandler;
import ca.uvic.csr.shrimp.usercontrols.DefaultViewAction;

/**
 * A panel containing buttons to invoke quick views. 
 * These buttons can also be drag & drop enabled. Dragging and dropping on the 
 * button will cause a new working set to be brought into Shrimp and the 
 * appropriate quick view applied.

 * @author Rob Lintern, Chris Callendar
 */
public class QuickViewPanel {
	
    private JPanel pnlQuickView;
    private JPanel pnlButtons;
    private TransferHandler transferHandler;
    private QuickViewManager manager;
    private ShrimpProject project;
    
    public QuickViewPanel(ShrimpProject project, TransferHandler transferHandler) {
        super();
        this.project = project;
        this.transferHandler = transferHandler;
        this.manager = project.getQuickViewManager();
        
        createGUI();
    }
    
    private void createGUI() {
        pnlQuickView = new JPanel(new BorderLayout(0, 0));
        pnlButtons = new JPanel();
        pnlButtons.setLayout(new BoxLayout(pnlButtons, BoxLayout.Y_AXIS));
        JLabel lbl = new JLabel(ResourceHandler.getIcon("icon_label_quick_views.gif"));
        //lbl.setBorder(BorderFactory.createEtchedBorder());
        pnlQuickView.add(lbl, BorderLayout.NORTH);
		pnlQuickView.add(pnlButtons, BorderLayout.CENTER);
		reloadQuickViewButtons();
		
		Action customAction = new AbstractAction("Customize the quick views for this project", 
										ResourceHandler.getIcon("icon_quick_view_custom_views.gif")) {
        	public void actionPerformed(ActionEvent e) {
				QuickViews.showQuickViewsDialog(ApplicationAccessor.getApplication().getParentFrame(), project.getQuickViewManager());
				reloadQuickViewButtons();
				pnlButtons.repaint();
				pnlButtons.revalidate();
        	}
        };

		JButton btn = createQuickViewButton(customAction, null);
		pnlQuickView.add(btn, BorderLayout.SOUTH);
    }

    private void reloadQuickViewButtons() {
    	pnlButtons.removeAll();
    	
    	// add the default view first (should always be valid)
    	QuickViewAction defaultAction = manager.getQuickViewAction(DefaultViewAction.ACTION_NAME);
    	if (defaultAction != null) {
    		JButton btn = createQuickViewButton(defaultAction, transferHandler);
    		pnlButtons.add(btn);
    	}

	    // now add the rest - but only the ones that are valid
    	QuickViewAction[] actions = manager.getSortedQuickViews();
		for (int i = 0; i < actions.length; i++) {
            QuickViewAction action = (QuickViewAction) actions[i];
            if ((action != defaultAction) && action.isValid()) {
            	JButton btn = createQuickViewButton(action, transferHandler);
            	pnlButtons.add(btn);
            }
        }
    }
    
    /**
     * Creates a quick view button and adds the transfer handler if not null.
     * This method does not add the button to a panel, it just returns it.
     * @param action the action for the button
     * @param handler the transfer handler, can be null
     * @return the created button
     */
    private JButton createQuickViewButton(Action action, TransferHandler handler) {
	    final JButton btnQuickView = new JButton(action);
	    btnQuickView.setToolTipText(btnQuickView.getText());
	    if (btnQuickView.getIcon() != null) {
	        // if the button has an icon then just show it without text
	        btnQuickView.setText(null);
	        btnQuickView.setIconTextGap(0);
	    } 
        btnQuickView.setMargin(new Insets(1,1,1,1));
        btnQuickView.setPreferredSize(new Dimension(56, 56));
        btnQuickView.setBackground(Color.WHITE);
        if (handler != null) {
            btnQuickView.setTransferHandler(handler);   
        }
        return btnQuickView;
    }
    
    public JPanel getQuickViewPanel() {
        return pnlQuickView;
    }
    
}
