package uk.ac.ebi.interpro.graphdraw;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.mylar.zest.layouts.dataStructures.InternalNode;
import org.eclipse.mylar.zest.layouts.dataStructures.InternalRelationship;

/**
 * Layout a hierarchical graph.
 * <p>
 * Layout similar to the algorithm implemented by
 * <a href="http://www.graphviz.org/">GraphViz, AT&T laboratories</a>.
 * </p>
 * <pre>
 * Usage:
 * Graph g=....
 * HierarchicalLayout layout = new HierarchicalLayout(g,HierarchicalLayout.PARENT_TOP);
 * layout.layout();
 * ...layout.getWidth()...layout.getHeight()...
 * </pre>
 * <p>
 * See SimpleGraphDraw.java for an example.
 * </p>
 * <p>
 * The following documentation assumes orientation is PARENT_TOP,
 * in which case the parent nodes are at the top and
 * a level is a horizontal group of nodes.
 * Normally the magic field values should be left at their default values.
 * </p>
 */
public class HierarchicalLayout {

// Public interface

    /** Parent nodes at top of graph layout */
    public final static int PARENT_TOP = 0;
    /** Parent nodes at left of graph layout */
    public final static int PARENT_LEFT = 1;


    /**
     * Construct a layout from Zest nodes and edges.
     * Assume a PARENT_TOP orientation
     * @param entitiesToLayout
     * @param relationshipsToConsider
     */
    public HierarchicalLayout(InternalNode[] entitiesToLayout,
			InternalRelationship[] relationshipsToConsider) {
		Map hnodes = new HashMap();

		for (int i = 0; i < entitiesToLayout.length; i++) {
			InternalNode node = entitiesToLayout[i];
			HierarchicalNode hnode = new HierarchicalNode(node, 50, 50); // TODO - how to set height and width (maybe based on text?)
			hnodes.put(node, hnode);
			hierarchicalGraph.nodes.add(hnode);
		}
		for (int i = 0; i < relationshipsToConsider.length; i++) {
			InternalRelationship edge = relationshipsToConsider[i];
			hierarchicalGraph.edges.add(new HierarchicalEdge(
					(HierarchicalNode) hnodes.get(edge.getSource()),
					(HierarchicalNode) hnodes.get(edge.getDestination()), edge));
		}
	}

/**
 * Compute layout. This method finally calls setLocation on all the nodes and
 * setRoute on all the edges.
 */

    public void layout() {
        for (Iterator i = hierarchicalGraph.nodes.iterator(); i.hasNext();) {
            HierarchicalNode n = (HierarchicalNode) i.next();
            findLevel(maxLevelSize, n, 0);
        }

        rationalise();

        for (Iterator i = levels.iterator(); i.hasNext();) {
            Level l = (Level) i.next();
            l.calcInitialPositions();

        }

        orderNodesInLevels();

        calcLevelLocations();

        int minStart = Integer.MAX_VALUE;

        for (Iterator i = levels.iterator(); i.hasNext();)
            minStart = Math.min(minStart, ((Level) i.next()).getStart());

        for (Iterator i = levels.iterator(); i.hasNext();)
            ((Level) i.next()).shiftLeft(minStart);

        storeLayout();
    }

/**
 * After calling layout() call getWidth and getHeight
 * <br/>
 * All nodes will be in this bounding box.
 * <br/>
 * 0&lt;node.x+/-node.getWidth/2&lt;layout.getWidth
 * <br/>
 * 0&lt;node.y+/-node.getHeight/2&lt;layout.getHeight
 * <br/>
 * Noting that x and y are the centres of the nodes.
 * All edge routes will also be in the bounding box.   
 *
 * @return width of layout
 */

    public int getWidth() {
        int maxWidth = 0;

        for (Iterator i = levels.iterator(); i.hasNext();)
            maxWidth = Math.max(maxWidth, ((Level) i.next()).getWidth());

        return maxWidth;

    }


/** See getWidth() */
    public int getHeight() {
        Level l = (Level) levels.get(levels.size() - 1);
        return l.location + l.depth / 2;
    }
//
// Magic constants.
// WARNING! Playing with these constants will mess with your head.
//

/** Ratio of maximum edge vertical distance to horizontal distance */
    public int edgeLengthHeightRatio = 6;
/** Number of passes up and down the levels to attempt to optimise node positions */
    public final int reorderIterations = 25;
/** Minimum gap between levels */
    public final int minLevelGap = 10;
/** Levels may be split if they have more than this number of nodes */
    public final int maxLevelSize = 100;
/** Edges running though levels will be allocated this much horizontal space */
    public int insertedEdgeWidth = 20;
/** Horizontal gap between nodes */
    public int withinLevelGap = 20;
//


// Internal implementation

// fields

    private StandardGraph hierarchicalGraph = new StandardGraph();
    private int orientation = PARENT_TOP;
    private ArrayList levels = new ArrayList();
    private Set originalEdges;

// classes

    private class HierarchicalNode implements Node {

        // Underlying node will be null for inserted nodes
    	InternalNode underlying;

        /**
         * Level on which node is located
         */
        Level level;

        /**
         * Location within level
         */
        int location;

        /**
         * Size of node expressed wrt level
         */
        int withinLevelSize;

        /**
         * Size of node expressed wrt level
         */
        int betweenLevelSize;

        /**
         * sequence code - used to stabilise graph
         */
        int code;

        /**
         * Create nodedata for node
         */
        HierarchicalNode(InternalNode node, int betweenLevelSize, int withinLevelSize) {
            underlying = node;
            this.betweenLevelSize = betweenLevelSize;
            this.withinLevelSize = withinLevelSize;
        }

        HierarchicalNode(int betweenLevelSize, int withinLevelSize) {
            this.betweenLevelSize = betweenLevelSize;
            this.withinLevelSize = withinLevelSize;            
        }


    }

    private class HierarchicalEdge implements Edge {

    	HierarchicalNode parent, child;

        //underlying will be null for constructed edges
        InternalRelationship underlying = null;

        //componentEdges will be empty for constructed edges
        List componentEdges = new ArrayList();

        public Node getParent() {
            return parent;
        }

        public Node getChild() {
            return child;
        }

        public HierarchicalEdge(HierarchicalNode parent, HierarchicalNode child, InternalRelationship edge) {
            this.parent = parent;
            this.child = child;
            this.underlying = edge;
        }

		public HierarchicalEdge(HierarchicalNode parent, HierarchicalNode child) {
            this.parent = parent;
            this.child = child;
		}
    }


    private class Level {

        int levelNumber;
        //int size;
        int location, depth;
        List nodes = new ArrayList();
        StandardGraph hierarchicalGraph;

        public Level(StandardGraph hierarchicalGraph, int levelNumber) {
            this.hierarchicalGraph = hierarchicalGraph;

            this.levelNumber = levelNumber;
        }

        void reorder(Level above, Level below) {

            for (int j = 0; j < nodes.size(); j++) {

                HierarchicalNode nj = (HierarchicalNode) nodes.get(j);

                double total = 0;
                int connected = 0;

                if (above != null)
                    for (int i = 0; i < above.nodes.size(); i++) {
                        HierarchicalNode ni = (HierarchicalNode) above.nodes.get(i);


                        if (hierarchicalGraph.connected(ni, nj)) {
                            connected++;
                            total += ni.location;
                        }
                    }

                if (below != null)
                    for (int i = 0; i < below.nodes.size(); i++) {
                        HierarchicalNode ni = (HierarchicalNode) below.nodes.get(i);

                        if (hierarchicalGraph.connected(nj, ni)) {
                            connected++;
                            total += ni.location;
                        }
                    }

                if (connected == 0) {
                    continue;
                    //throw new RuntimeException("No connected Nodes");
                } else {
                    total /= connected;
                }

                nj.location = (int) total;
            }


            while (true) {

                Collections.sort(nodes, nodeLayoutComparator);

                boolean foundOverlap = false;
                for (int i = 1; i < nodes.size(); i++) {
                    HierarchicalNode a = (HierarchicalNode) nodes.get(i - 1);
                    HierarchicalNode b = (HierarchicalNode) nodes.get(i);

                    int overlap = minLevelGap + (a.location + a.withinLevelSize / 2) - (b.location - b.withinLevelSize / 2);
                    if (overlap > 0) {
                        foundOverlap = true;
                        a.location = a.location - overlap / 2 - 1;
                        b.location = b.location + overlap / 2 + 1;
                    }
                }
                if (!foundOverlap) break;
            }

/*    		excess=0;
if (excess>((Node)nodes.get(0)).x) {excess=((Node)nodes.get(0)).x;}
shiftLeft(excess);*/

        }


        void calcInitialPositions() {

            int width = 0;
            for (int i = 0; i < nodes.size(); i++) {
                HierarchicalNode n = (HierarchicalNode) nodes.get(i);

                n.location = width + n.withinLevelSize / 2;
                width += n.withinLevelSize + withinLevelGap;
            }


        }

        void shiftLeft(int delta) {
            for (int i = 0; i < nodes.size(); i++) {
                ((HierarchicalNode) nodes.get(i)).location -= delta;
            }
        }

        void setDepth(int depth) {
            this.depth = depth;

        }

        void setLocation(int location) {
            this.location = location;
        }

        int getWidth() {
            final HierarchicalNode nd = ((HierarchicalNode) nodes.get(nodes.size() - 1));
            return nd.location + nd.withinLevelSize / 2;
        }


        int getStart() {
            final HierarchicalNode nd = ((HierarchicalNode) nodes.get(0));
            return nd.location - nd.withinLevelSize / 2;
        }


    }

    private static final Comparator nodeLayoutComparator = new Comparator() {
        // Classcast exception if not NodeLayoutData
        public int compare(Object o1, Object o2) {
            return ((HierarchicalNode) o1).location - ((HierarchicalNode) o2).location;
        }

    };

// methods

    private int findLevel(int maxLevelSize, HierarchicalNode node, int nAttempts) {
    	
    	final int MAX_ATTEMPTS = 100;

        if (node.level != null) return node.level.levelNumber;

        int maxParentLevel = -1;

        Set parents = hierarchicalGraph.parents(node);

        for (Iterator i = parents.iterator(); i.hasNext();) {
            HierarchicalNode parent = (HierarchicalNode) i.next();
            if (parent != null && nAttempts < MAX_ATTEMPTS) {
            	int l = findLevel(maxLevelSize, parent, nAttempts+1);
            	if (l > maxParentLevel) maxParentLevel = l;
            }	
        }


        int levelNumber = maxParentLevel + 1;

        while (true) {

            while (levelNumber >= levels.size()) levels.add(new Level(hierarchicalGraph, levels.size()));

            if (((Level) levels.get(levelNumber)).nodes.size() < maxLevelSize) break;

            levelNumber++;
        }

        node.level = (Level) levels.get(levelNumber);

        node.level.nodes.add(node);

        return levelNumber;
    }


    private void rationalise(HierarchicalEdge e, StandardGraph g) {
        int parentLevel = ((HierarchicalNode) e.getParent()).level.levelNumber;
        int childLevel = ((HierarchicalNode) e.getChild()).level.levelNumber;

        if (parentLevel < childLevel - 1) {
            //System.out.println("Rationalise "+parentLevel+" "+childLevel);
            HierarchicalNode a = (HierarchicalNode) e.getParent();
            for (int i = parentLevel + 1; i <= childLevel; i++) {

                HierarchicalNode b;
                if (i == childLevel) {
                    b = (HierarchicalNode) e.getChild();
                } else {
                    b = new HierarchicalNode(-1, insertedEdgeWidth);
                    b.level = (Level) levels.get(i);
                    b.level.nodes.add(b);

                }
                HierarchicalEdge insertedEdge = new HierarchicalEdge(a, b);
                g.edges.add(insertedEdge);
                g.nodes.add(b);
                e.componentEdges.add(insertedEdge);

                a = b;
            }

        } else {
            e.componentEdges.add(e);
            g.edges.add(e);
        }
    }

    private void rationalise() {
        originalEdges = new HashSet(hierarchicalGraph.edges);

        hierarchicalGraph.edges.clear();

        for (Iterator i = originalEdges.iterator(); i.hasNext();) {
            HierarchicalEdge e = (HierarchicalEdge) i.next();
            rationalise(e, hierarchicalGraph);
        }
    }

    private void orderNodesInLevels() {
        for (int j = 0; j < reorderIterations; j++) {

            int s = levels.size();

            for (int i = 0; i < s; i++) {
                Level p = (i == 0) ? null : (Level) levels.get(i - 1);
                Level l = (Level) levels.get(i);
                Level n = (i == s - 1) ? null : (Level) levels.get(i + 1);
                l.reorder(p, n);
            }

            for (int i = s - 1; i >= 0; i--) {
                Level p = (i == 0) ? null : (Level) levels.get(i - 1);
                Level l = (Level) levels.get(i);
                Level n = (i == s - 1) ? null : (Level) levels.get(i + 1);
                l.reorder(p, n);
            }

        }
    }


    private void calcLevelLocations() {
        int height = 0;

        Level p = null;

        for (Iterator i = levels.iterator(); i.hasNext();) {

            Level l = (Level) i.next();

            int maxLength = 0;

            // Calculate maximum edge length

            if (p != null) {
                for (Iterator i2 = l.nodes.iterator(); i2.hasNext();) {
                    HierarchicalNode n1 = (HierarchicalNode) i2.next();
                    for (Iterator i3 = p.nodes.iterator(); i3.hasNext();) {
                        HierarchicalNode n2 = (HierarchicalNode) i3.next();
                        if (hierarchicalGraph.connected(n1, n2)) {
                            maxLength = Math.max(maxLength, Math.abs(n1.location - n2.location));
                        }
                    }
                }
                height += Math.max(minLevelGap, maxLength / edgeLengthHeightRatio);
            }


            int maxHeight = 0;

            for (Iterator i2 = l.nodes.iterator(); i2.hasNext();) {
                maxHeight = Math.max(maxHeight, ((HierarchicalNode) i2.next()).betweenLevelSize);
            }

            l.setDepth(maxHeight);

            height += l.depth / 2;

            l.setLocation(height);

            height += maxHeight;

            p = l;
        }
    }

    private void storeLayout() {
        for (Iterator it = hierarchicalGraph.nodes.iterator(); it.hasNext();) {
            HierarchicalNode n = (HierarchicalNode) it.next();
            if (n.underlying == null) continue;
            n.underlying.setInternalLocation(n.location, n.level.location);
            n.underlying.setInternalSize(n.betweenLevelSize, n.withinLevelSize);
        }


        for (Iterator it = originalEdges.iterator(); it.hasNext();) {
            HierarchicalEdge e = (HierarchicalEdge) it.next();

            for (int i = 0; i < e.componentEdges.size(); i++) {

                HierarchicalEdge edge = (HierarchicalEdge) e.componentEdges.get(i);
                HierarchicalNode parent = (HierarchicalNode) edge.getParent();
                HierarchicalNode child = (HierarchicalNode) edge.getChild();

                int parentLocation = parent.location;
                int childLocation = child.location;

                int levelParent = parent.level.location + parent.level.depth / 2;
                int levelChild = child.level.location - child.level.depth / 2;

                int levelCentre = (levelParent + levelChild) / 2;

                int nodeParent = parent.level.location + parent.betweenLevelSize / 2;
                int nodeChild = child.level.location - child.betweenLevelSize / 2;

                if (orientation == PARENT_TOP) {
                	e.underlying.addBendPoint(parentLocation, nodeParent);
                	e.underlying.addBendPoint(parentLocation, levelParent);
                	e.underlying.addBendPoint(parentLocation, levelCentre, true /*control*/);
                	e.underlying.addBendPoint(childLocation, levelCentre, true /*control*/);
                	e.underlying.addBendPoint(childLocation, levelChild);
                	e.underlying.addBendPoint(childLocation, nodeChild);
                }
            }
        }
    }

}
